#
# Module for the handling of stochastic distributions.
#
######################################################################
# The two following procedures (Law_Browser and UpdateLawParams)
# allow to greate a Browser for some distribution law.
# The description itself is retuned in a global array named LawParams. 
# The "fields" of this array are:
# - name	the name of the distribution
# - mean	the mean of the distribution
# - alpha	the "alpha" parameter of Pareto distributions
######################################################################
# Law_Browser -
# Creates a frame with a menu allowing to choose a distribution.
# The frame is attached to an existing one passed as a parameter.
# The name is returned as result.
#
# Parameter:
# - win		the window (frame) into which to pack the browser
# - value	the name of the array keeping the parameters
#
proc Law_Browser { win value } {
    global LawDesc
    upvar $value Params

    # puts "Browsing with law [array get Params]"

    set f [frame $win.law]
    menubutton $f.choice -text "Law:" -width 5 -underline 0 \
	-menu $f.choice.menu -anchor nw

    set m [menu $f.choice.menu]
    foreach name [array names LawDesc] {
	set Name [UpperCase $name]
	$m add command -label $Name \
	    -command "UpdateLawParams $f.namelabel $f.params $value $name" \
	    -underline 0
    }

    label $f.namelabel -width 12 -pady 3 -anchor w \
	-text [UpperCase $Params(name)]
    pack $f.choice $f.namelabel -side left -anchor n

    # Create the default parameter area and initialize it
    frame $f.params
    UpdateLawParams $f.namelabel $f.params $value $Params(name)

    # Do not pack the frame: the caller will do it.
    # Return the frame name.
    return $f

}

######################################################################
# UpdateLawParams -
# The procedure that handles the changes in the law menu.
# It creates and packs various frames inside the frame passed as
# a parameter, according to the parameters of the current law.
#
# Parameters:
# - labelwin	the label containing the name to be updated
# - win		the frame containing the parameters of the law
# - value	the name of array containing  the parameters
# - name	the name of the just chosen law
#
proc UpdateLawParams { labelwin win value name } {
    global LawDesc
    upvar $value Params

    set labelname [UpperCase $name]
    set Params(name) $name
    $labelwin configure -text $labelname

    # puts "Updating with law [array get Params]"

    # get the generic description of the law
    array set TheLaw $LawDesc($name)

    # clean up the frame and forget about the components
    foreach child [winfo children $win] {
	pack forget $child
    }

    # create the zones corresponding to the entries of the array, except
    # "name".
    unset TheLaw(name)
    set zones [array names TheLaw] 
    
    foreach zone $zones {
	if ![winfo exists $win.$zone] { 
	    frame $win.$zone
	    if ![info exists Params($zone)] {
		set Params($zone) $TheLaw($zone)
	    }
	} else {
	    # clean up the frame
	    foreach child [winfo children $win.$zone] {
		destroy $child
	    }
	}

	label $win.$zone.label -text [UpperCase $zone] -width 15 -anchor w
	entry $win.$zone.field -textvar ${value}($zone) \
	    -width 10 -relief sunken
	pack $win.$zone.label $win.$zone.field -side left 
	pack $win.$zone -side top
    }

    pack $win
}

######################################################################
# write_law -
# Procedure to write out some probability distribution in the ERS
# format (for EGL PNL etc)
#
# Parameters:
# - outfile	the stream into which to write
# - lawdesc	the list of parameters of the distribution
#
proc write_law { outfile lawdesc } {

    array set TheLaw $lawdesc

    switch $TheLaw(name) {
	bernoulli {
	    puts -nonewline $outfile "B [expr double($TheLaw(proba))]"
	}
	constant {
	    puts -nonewline $outfile "C [expr double($TheLaw(value))]"
	}
	exponential {
	    puts -nonewline $outfile "E [expr double($TheLaw(mean))]"
	}
	geometric {
	    puts -nonewline $outfile "G [expr double($TheLaw(mean))]"
	}
	pareto {
	    puts -nonewline $outfile \
		"p [expr double($TheLaw(alpha))] [expr double($TheLaw(beta))]"
	}
	uniform {
	    puts -nonewline $outfile \
		"U [expr double($TheLaw(low))] [expr double($TheLaw(high))]"
	}
	default {
	    ReportError "Export Error" "Don't know distribution $TheLaw(name)"
	}
    }
}

######################################################################
# Is_Numeric_Law -
# Boolean function telling whether the law contains only proper numbers
# and no symbols or anything else.
#
# Parameters:
# - lawdesc	the list of parameters of the distribution
#
proc Is_Numeric_Law { lawdesc } {

    array set TheLaw $lawdesc

    switch $TheLaw(name) {
	bernoulli {
	    return [isReal $TheLaw(proba)]
	}
	constant {
	    return [isReal $TheLaw(value)]
	}
	exponential -
	geometric {
	    return [isReal $TheLaw(mean)]
	}
	pareto {
	    return [expr [isReal $TheLaw(alpha)] \
			&& [isReal $TheLaw(beta)]]
	}
	uniform {
	    return [expr [isReal $TheLaw(low)] \
			&& [isReal $TheLaw(high)]]
	}
	default {
	    ReportError "Export Error" "Don't know distribution $TheLaw(name)"
	}
    }
}

######################################################################
# SetLaw -
# Procedure to put into an array the parameters of the current law.
# Only the fields relevant to this law are extracted from the richer
# LawParams array.
## DOES NOT WORK, but I don't understand why...
#
# Parameters:
# - arr		the array to instanciate
# - index	the index of the array at which to put the law array
#
proc SetLaw { arr index } {
    global LawDesc LawParams
    global a
    upvar $arr a

    if [info exists a($index)] {
	unset a($index)
    }
    array set TheLaw "$LawDesc($LawParams(name))"
    foreach field [array names TheLaw] {
	set l($field) $LawParams($field)
    }
    array set a($index) [array get l]
    # puts "$arr set to [array get a]"
}

######################################################################
# Is_Zero_Time - 
# Boolean procedure telling whether a time distribution is actually
# constant with value zero.
#
# Parameters:
# - law		the law to test
#
proc Is_Zero_Time { law } {
    upvar $law Law

    # test law parameters, but try to protect against bad numbers
    switch $Law(name) {
	"constant" { 
	    set condition "set res \[expr $Law(value) == 0.0\]"
	}
	"erlang" -
	"exponential" -
	"geometric" { 
	    set condition "set res \[expr $Law(mean) == 0.0\]"
	}
	"uniform" { 
	    set condition "set res \[expr ( $Law(low) == 0.0 ) && ( $Law(low) == $Law(high) )\]"
	}
	default { 
	    set condition "set res 0"
	}
    }

    if [catch $condition err] {
	return 0
    } else {
	return $res
    }
}

