#
# The I/O Module for gred
#
# This module contains the procedures to read and write files
# - save/load in the graph format
# - save as exclusion graph
#
######################################################################
# write_graph -
# Saves the current Interaction Graph in the "gred" format.
#
# Parameters:
# - win		the canvas containing the chain to save
# - file_name	the name of the file in which to save.
#
proc write_graph { win file_name } {
    global node_list arc_list
    global weight
    global DEBUG

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Saving graph in file $file_name..."
    }

    # renumbering all nodes
    set curnode 1
    foreach node $node_list {
	set no_number([lindex $node 0]) $curnode
	incr curnode 1
    }

    foreach node $node_list {
	set node_id [lindex $node 0]
	puts -nonewline $outfile "Node $no_number($node_id) "
	puts -nonewline $outfile [$win coords $node_id]
# 	regsub -all {[ ]} $name($node_id) "_WHITE_" the_name
# 	puts -nonewline $outfile " \"$the_name\""
	puts $outfile " $weight($node_id)"
    }

    foreach link $arc_list {
	set link_id [lindex $link 0]
	set orig_id [lindex $link 1]
	set dest_id [lindex $link 2]
	puts -nonewline $outfile "Arc "
	puts -nonewline $outfile "$no_number($orig_id) $no_number($dest_id) "
	puts $outfile [$win coords $link_id]
    }

    close $outfile
    if $DEBUG {
	puts " ...Done"
    }
}

proc read_graph { win file_name merge { Xoff 0 } { Yoff 0 } } {
    global name weight object_distance
    global node_list arc_list
    global DEBUG

    if [catch {open $file_name r} infile] {
	puts "Could not open file $file_name for reading"
	return 
    }

    if !$merge {
	Clear_Graph $win
    }

    if $DEBUG {
	puts -nonewline "Reading graph from $file_name..."
    }

    # convert offset into pixels
    set xoff [winfo pixels . ${Xoff}m]
    set yoff [winfo pixels . ${Yoff}m]

    while { [gets $infile line] >= 0 } {
	set args [split $line]
	switch [lindex $args 0] {
	    Node { 
		set xpos [expr $xoff+([lindex $args 2]+[lindex $args 4])/2]
		set ypos [expr $yoff+([lindex $args 3]+[lindex $args 5])/2]
		set newid [make_new_node $win $xpos $ypos]
# 		set the_name [string trim [lindex $args 6] \"]
# 		regsub {_WHITE_} $the_name " " name($newid)
#		Display_Name $newid
		set weight($newid) [lindex $args 6]
		set no_id([lindex $args 1]) $newid
	    }
	    Arc { 
		set points [Shift_Line_Coords \
				[lrange $args 3 [llength $args]] \
				$xoff $yoff]
		set newid [eval {$win create line} $points -tags ARC]
		set newlink [list $newid $no_id([lindex $args 1]) \
				 $no_id([lindex $args 2])]
		lappend arc_list $newlink
	    }
	}
    }

    # adjusting the size of the canvas
    set bbox [$win bbox NODE ARC]
    Resize_Graph $win [expr [lindex $bbox 2]+$object_distance] \
	[expr [lindex $bbox 3]+$object_distance]
    
    close $infile
    puts " ...Done"
}

proc Save_Graph { win } {
    global file_name state

    set types {
	{{Graph File}       {.gr .gred .graph}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getSaveFile -defaultextension gr \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension gred -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetSaveFile $options]

    if { $file_name != "" } {
	write_graph $win $file_name
	set state($win) saved
	return 1
    } else {
	return 0
    }
}

proc Browse_File { } {
    global file_name 

    set types {
	{{Graph File}       {.gr .gred .graph}  }
	{{All Files}        *             }
    }

#     set file_name [tk_getOpenFile -defaultextension pned \
# 		       -initialfile [file tail $file_name] \
# 		       -filetypes $types -parent .]
    set options "-defaultextension gred -initialfile {[file tail $file_name]} -filetypes {$types} -parent ."

    set file_name [GetOpenFile $options]

    return $file_name
		  
}

proc Load_Graph { win } {
    global file_name GeomParams

    if [catch {toplevel .params}] { return }

    wm title .params "Load"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar file_name -width 20 \
	-relief sunken

    frame .params.offset
    label .params.offset.label -text "Offset (mm)" -width 20 -anchor w
    label .params.offset.xlabel -text "X=" -width 3 -anchor w
    entry .params.offset.xoffset -textvar GeomParams(xoffset) \
	-width 4 -relief sunken
    label .params.offset.ylabel -text "Y=" -width 3 -anchor w
    entry .params.offset.yoffset -textvar GeomParams(yoffset) \
	-width 4 -relief sunken
    pack .params.offset.label .params.offset.xlabel .params.offset.xoffset \
	.params.offset.ylabel .params.offset.yoffset -side left 

    frame .params.actions
    button .params.actions.browse -text "Browse" \
	-underline 0 \
	-command Browse_File
    button .params.actions.ok -text "Load" \
	-underline 0 \
	-command "destroy .params\n read_graph $win \$file_name 0 \$GeomParams(xoffset) \$GeomParams(yoffset) "
    button .params.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy .params\n read_graph $win \$file_name 1 \$GeomParams(xoffset) \$GeomParams(yoffset)"
    button .params.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .params\n array set GeomParams {\[array get GeomParams\]}"

    pack .params.actions.browse .params.actions.ok \
	.params.actions.merge .params.actions.cancel \
	-side left

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename.label -side left
    pack .params.filename.field -side left -expand true -fill x
    pack .params.filename -side top -fill x
    pack .params.offset -side top
    pack .params.actions -side top

    grab .params

}

######################################################################
# Procedures for exporting the graph in other formats than the gred
# formats.
######################################################################
# write_matrix -
# Exports the current graph as an adjacency matrix
#
# Parameters:
# - file_name	the name of the file in which to write the result
# - mode	the representation of the matrix: one of sparse/full
#
proc write_matrix { file_name { mode sparse } } {
    global node_list arc_list weight
    global DEBUG

    set SPA [expr 0 == [string compare $mode sparse]]

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating Adjacency matrix (mode $mode)..."
    }

    # numbering the nodes
    set curnodenumber [llength $node_list]
    for { set i 0 } { $i < $curnodenumber } { incr i 1 } {
	set no_number([lindex [lindex $node_list $i] 0]) $i
    }

    # type of representation
    if $SPA {
	puts $outfile "sp"
    }

    # number of nodes
    puts $outfile $curnodenumber

    # No need to compute probas by normalizing the weights. 
    # Matrix programs should handle that themselves
    foreach node $node_list {
	set i [lindex $node 0]
	puts -nonewline $outfile " $weight($i)"
    }
    puts $outfile ""

    # adjacency matrix
    foreach node $node_list {
	set orig [lindex $node 0]

	# scan arcs looking for neighbors. A note is ALWAYS neighbot of 
	# itself.
	set neighbors $orig
	foreach arc $arc_list {
	    if { $orig == [lindex $arc 1] } {
		set neighbors [lappend neighbors [lindex $arc 2]]
	    } elseif { $orig == [lindex $arc 2] } {
		set neighbors [lappend neighbors [lindex $arc 1]]
	    }
	}
	if $DEBUG {
	    puts "Node $orig: [llength $neighbors] links ($neighbors)"
	}
	
	if $SPA {
	    foreach ngb $neighbors {
		puts $outfile "$no_number($orig) $no_number($ngb)"
	    }
	} else {
	    foreach othernode $node_list {
		set dest [lindex $othernode 0]
		if [expr {-1 != [lsearch $neighbors $dest]}] {
		    puts -nonewline $outfile " 1"
		} else {
		    puts -nonewline $outfile " 0"
		}
	    }
	    puts $outfile ""
	}
    }
    close $outfile

    if $DEBUG {
	puts " ...Done"
    }

    return 1
}

proc gen_matrix { win mode } {
    global gen_file_name

    set types {
	{{Adjacency Matrix}       {.adj}  }
	{{All Files}        *             }
    }

#     set gen_file_name [tk_getSaveFile -defaultextension adj \
# 			   -filetypes $types -parent .]
    set options "-defaultextension adj -filetypes {$types} -parent ."

    set gen_file_name [GetSaveFile $options]

    if { $gen_file_name != "" } {
	write_matrix $gen_file_name $mode
    }

}

proc write_maxplus_matrices { file_name format prefix} {
    global node_list arc_list weight
    global DEBUG

    set SCI [expr 0 == [string compare $format scilab]]
    set SPA [expr 0 == [string compare $format sparse]]

    set outfile [open $file_name w]

    if $DEBUG {
	puts -nonewline "Generating the set of (max,+) matrices...";
    }

    # numbering the nodes
    set curnodenumber [llength $node_list]
    for { set i 0 } { $i < $curnodenumber } { incr i 1 } {
	set no_number([lindex [lindex $node_list $i] 0]) $i
    }

    # number of matrices and their size
    if !$SCI {
	puts $outfile $curnodenumber
	puts $outfile $curnodenumber
    }

    if $SCI {
	puts -nonewline $outfile "Proba_$prefix = \["
    }
    # No need to compute probas by normalizing the weights. 
    # Matrix programs should handle that themselves
    foreach node $node_list {
	set i [lindex $node 0]
	puts -nonewline $outfile " $weight($i)"
    }
    if $SCI {
	puts $outfile " \]"
    } else {
	puts $outfile ""
    }

    # now the matrices
    foreach node $node_list {
	set the_node [lindex $node 0]

	# scan arcs looking for neighbors 
	set neighbors $the_node
	foreach arc $arc_list {
	    if { $the_node == [lindex $arc 1] } {
		set neighbors [lappend neighbors [lindex $arc 2]]
	    } elseif { $the_node == [lindex $arc 2] } {
		set neighbors [lappend neighbors [lindex $arc 1]]
	    }
	}
	if $DEBUG {
	    puts "Node $the_node: [llength $neighbors] links ($neighbors)"
	}

	if $SCI {
	    set matrixnum [lsearch $node_list $node]
	    puts -nonewline $outfile "$prefix$matrixnum = \#(\[\n\t"
	}

	if $SPA {
	    puts $outfile "sp"
	}
	for { set i 0 } { $i < $curnodenumber } { incr i 1 } {
	    set orig [lindex [lindex $node_list $i] 0]
	    for { set j 0 } { $j < $curnodenumber } { incr j 1 } {
		set dest [lindex [lindex $node_list $j] 0]
		if [expr ( $orig == $the_node ) \
			&& ( -1 != [lsearch $neighbors $dest] )] {
		    if $SPA {
			puts $outfile "$i $j 1"
		    } else {
			puts -nonewline $outfile "  1  "
		    }
		} elseif { $i == $j } {
		    if $SPA {
			puts $outfile "$i $j 0"
		    } else {
			puts -nonewline $outfile "  0  "
		    }
		} else {
		    if $SCI {
			puts -nonewline $outfile "  %0 "
		    } elseif !$SPA {
			puts -nonewline $outfile " eps "
		    }
		}
	    }
	    if $SCI {
		if { $i < [expr $curnodenumber-1] } {
		    puts -nonewline $outfile " ;\n\t"
		} else {
		    puts -nonewline $outfile " ])\n"
		}
	    } elseif !$SPA {
		puts $outfile ""
	    }
	}
	if $SPA {
	    puts $outfile "end"
	}
	puts $outfile ""
    }
    close $outfile

    if $DEBUG {
	puts " ...Done"
    }

    return 1
}

proc gen_maxplus_matrices { win } {
    global gen_file_name matrix_format matrix_prefix

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken
    
    frame .params.format 
    label .params.format.label -text "Format" -width 15 -anchor w
    radiobutton .params.format.sparse -text "Sparse" \
	-variable matrix_format -value sparse
    radiobutton .params.format.standard -text "Standard" \
	-variable matrix_format -value normal
    radiobutton .params.format.scilab -text "Scilab" \
	-variable matrix_format -value scilab

    frame .params.prefix
    label .params.prefix.label -text "Matrix Name" -width 15 -anchor w
    entry .params.prefix.field -textvar matrix_prefix -width 20 -relief sunken

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\write_maxplus_matrices \$gen_file_name \$matrix_format \$matrix_prefix"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename.label .params.filename.field -side left
    pack .params.format.label .params.format.sparse .params.format.standard \
	.params.format.scilab -side left -fill x
    pack .params.prefix.label .params.prefix.field -side left -fill x
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename .params.format .params.prefix -side top -fill x
    pack .params.actions -side top

}

proc write_tetris { file_name opt } {
    global DEBUG KEEP

    set TMP_GRAPH_FILE "/tmp/gred[pid].graph"

    if $DEBUG {
	puts -nonewline "Generating Tetris set with option $opt..."
    }

    switch $opt {
	standard {
	    set outfile [open $file_name w]
	    Gen_Canonical_Tetris $outfile
	    close $outfile
	}
	edges {
	    if [write_matrix $TMP_GRAPH_FILE sparse] {
		exec gr2tetris -f $TMP_GRAPH_FILE -o $file_name
	    }
	}
	triangles {
	    if [write_matrix $TMP_GRAPH_FILE sparse] {
		exec gr2tetris -f $TMP_GRAPH_FILE -maxsize 3 -o $file_name
	    }
	}
	triangles_greedy {
	    if [write_matrix $TMP_GRAPH_FILE sparse] {
		exec gr2tetris -f $TMP_GRAPH_FILE -maxsize 3 -greedy \
		    -o $file_name
	    }
	}
    }

    if $DEBUG { 
	puts " ...Done"
    }

    if !$KEEP {
	RemoveFile $TMP_GRAPH_FILE
    }
}

######################################################################
# Gen_Canonical_Tetris -
# Create the "canonical" set of binary pieces corresponding to the arcs.
# See Brilman, thesis, p. 26.
# This has been superseded by the use of gr2tetris, in principle...
#
# Parameter:
# - outfile	the name of the stream into which to write
#
proc Gen_Canonical_Tetris { outfile } {
    global node_list arc_list weight color
    global DEBUG

    if $DEBUG {
	puts "doing the obsolete in-site generation..."
    }

    # numbering the nodes
    set curnodenumber [llength $node_list]
    for { set i 0 } { $i < $curnodenumber } { incr i 1 } {
	set no_number([lindex [lindex $node_list $i] 0]) $i
    }
    
    # number of resources
    set curarcnumber [llength $arc_list]
    puts $outfile "Columns $curarcnumber"
    puts $outfile ""

    # making colors
    Create_Colors $curnodenumber

    set num 1

    # now making the pieces
    foreach node $node_list {
	set the_node [lindex $node 0]

	# scan arcs looking for neighbors, stored as the arc number
	set neighbors ""
	for { set k 0 } { $k < $curarcnumber } { incr k 1 } {
	    set arc [lindex $arc_list $k]
	    if [expr { $the_node == [lindex $arc 1] } \
		    || { $the_node == [lindex $arc 2] } ] {
		set neighbors [lappend neighbors $k]
	    }
	}
	if $DEBUG {
	    puts "Node $the_node: [llength $neighbors] links ($neighbors)"
	}

	# construction of the profiles (after reinitialization)
	for { set k 0 } { $k < $curarcnumber } { incr k 1 } {
	    set LoProf($k) eps
	    set HiProf($k) eps
	}
	foreach k $neighbors {
	    set LoProf($k) 0
	    set HiProf($k) 0
	}

	# output of the piece
	puts $outfile "Piece $num"
	puts $outfile "Weight $weight($the_node)"
	puts $outfile "Color $color($num)"
	puts $outfile "LoProf [array get LoProf]"
	puts $outfile "HiProf [array get HiProf]"

	puts $outfile ""

	incr num 1
    }

    return 1
}

proc gen_tetris { win } {
    global gen_file_name
    global algo

    if [catch {toplevel .params}] { return }

    wm title .params "Generation Parameters"

    frame .params.filename
    label .params.filename.label -text "File Name" -width 15 -anchor w
    entry .params.filename.field -textvar gen_file_name -width 20 \
	-relief sunken
    
    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n\write_tetris \$gen_file_name \$algo"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"

    frame .params.algo
    radiobutton .params.algo.standard  -text "Standard"   -variable algo \
	-value standard
    radiobutton .params.algo.canonical -text "Canonical"  -variable algo \
	-value edges
    radiobutton .params.algo.triangles -text "Triangles"  -variable algo \
	-value triangles
    radiobutton .params.algo.trianggr  -text "Triangles Greedy" \
	-variable algo \
	-value triangles_greedy

    bind .params <Return> {.params.actions.ok invoke}
    bind .params <Escape> {.params.actions.cancel invoke}

    pack .params.filename.label .params.filename.field -side left
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.filename -side top -fill x
    pack .params.algo.standard .params.algo.canonical .params.algo.triangles \
	.params.algo.trianggr -side left -fill x
    pack .params.algo -side top -fill x
    pack .params.actions -side top

}

proc Gen_Postscript { win } {
    global file_name

    set types {
	{{Postscript File}       {.ps .eps .epsi}  }
	{{All Files}        *             }
    }
    
#     set ps_file_name [tk_getSaveFile -defaultextension ps \
# 			  -initialfile [Make_File_Name $file_name ps] \
# 			  -filetypes $types -parent .]
    set options "-defaultextension ps -initialfile {[Make_File_Name $file_name ps]} -filetypes {$types} -parent ."

    set ps_file_name [GetSaveFile $options]

    if { $ps_file_name != "" } {
	print_graph $win $ps_file_name 
    }

}

######################################################################
# Num2Col -
# Function converting an angle (measured in radians) into an integer
# in the hexadecimal format, between 00 and ff.
#
# Parameters:
# - Alpha	angle
#
proc Num2Col { Alpha } {
    global PI
    global DEBUG 


    set res [format "%02x" [expr int(floor(256*3*$Alpha/$PI))]]
    if { [string compare $res "100" ] == 0 } {
	set res "ff"
    }

    if $DEBUG { puts "Col: $Alpha -> $res " }

    return $res
}

######################################################################
# Create_Colors -
# Procedure to create a number of different colors using the "color
# circle". The colors are stored in the global array "color".
#
# Parameters:
# - number	the number of different colors to generate.
#
proc Create_Colors { number } {
    global PI color

    for { set i 0 } { $i < $number } { incr i 1 } {
	set Alpha [expr 2.0*$PI*$i/double($number)]

	if [expr $Alpha < $PI/3.0] {
	    set red "ff"
	    set green "00"
	    set blue [Num2Col $Alpha]
	} elseif [expr $Alpha < 2.0*$PI/3.0] {
	    set Alpha [expr $Alpha-$PI/3.0]
	    set red "ff"
	    set green [Num2Col $Alpha]
	    set blue "00"
	} elseif [expr $Alpha < $PI] {
	    set Alpha [expr $Alpha-2.0*$PI/3.0]
	    set red [Num2Col $Alpha]
	    set green "ff"
	    set blue "00"
	} elseif [expr $Alpha < 4.0*$PI/3.0] {
	    set Alpha [expr $Alpha-3.0*$PI/3.0]
	    set red "00"
	    set green "ff"
	    set blue [Num2Col $Alpha]
	} elseif [expr $Alpha < 5.0*$PI/3.0] {
	    set Alpha [expr $Alpha-4.0*$PI/3.0]
	    set red "00"
	    set green [Num2Col $Alpha]
	    set blue "ff"
	} else {
	    set Alpha [expr $Alpha-5.0*$PI/3.0]
	    set red [Num2Col $Alpha]
	    set green "00"
	    set blue "ff"
	}

	set Col "\#"
	append Col $red $green $blue
	puts $Col
	set color([expr $i+1]) $Col
    }
}
