#
# Module for the management of edit/select/delete modes (and others)
# 
######################################################################
# Edit_Object_Attributes -
# Open the window for editing some given object's attributes.
# Calls the proper type-specific editing procedure. 
# Recognizes attribute-less objects (marks,text) replaces them with their
# owner object.
# Coordinates (Xpos,Ypos) are passed to the procedure. They are NOT 
# used to find the object, but to position the window close to the object.
#
# Parameters:
# - win		the window where this is happening
# - obj_id	the id of the object
# - Xpos	a x coordinate
# - Ypos	a y coordinate
#
proc Edit_Object_Attributes { win obj_id Xpos Ypos } {
    global DEBUG 

    if $DEBUG {
	puts "Editing Attributes of $obj_id"
    }

    if [Is_Type $obj_id NODE] {
	Edit_Node_Attributes $win $Xpos $Ypos $obj_id
    } elseif [Is_Type $obj_id ARC] {
	Edit_Arc_Attributes $win $Xpos $Ypos $obj_id
    } elseif [Is_Type $obj_id TEXT] {
	# Either a name or a weight. Find its object id.
	set objtags  [$win gettags $obj_id]
	foreach tag $objtags {
	    if [expr [string match "Tx*" $tag] \
		    || [string match "Pr*" $tag] ] {
		set new_obj_id [string range $tag 2 end]
	    }
	}
	if [info exists new_obj_id] {
	    Edit_Object_Attributes $win $new_obj_id $Xpos $Ypos 
	}
    }
}

######################################################################
# Edit_Current_Object_Attributes -
# Open the window for editing the **current** object's attributes.
# Coordinates (Xpos,Ypos) are passed to the procedure. They are NOT 
# used to find the object, but to position the window close to the object.
#
# Parameters:
# - win		the window where this is happening
# - Xpos	a x coordinate
# - Ypos	a y coordinate
#
proc Edit_Current_Object_Attributes { win Xpos Ypos } {
    global DEBUG

    set cur_id [$win find withtag current]

    if { $cur_id != "" } { 
	Edit_Object_Attributes $win $cur_id $Xpos $Ypos
    }
}

######################################################################
# Edit_Node_Attributes -
# Open the window for editing the attributes of a node.
# Coordinates (Xpos,Ypos) are passed to the procedure. They are NOT 
# used to find the object, but to position the window close to the object.
#
# Parameters:
# - win		the window where this is happening
# - Xpos	a x coordinate
# - Ypos	a y coordinate
# - the_node_id	id of the node to edit
#
proc Edit_Node_Attributes { win Xpos Ypos the_node_id } {
    global name

    # for the time being, no attributes are editable for nodes!
    return

    if [catch {toplevel .att}] { return }

    wm title .att "Attributes of node $the_node_id"

    frame .att.name
    label .att.name.label -text "Name" -width 15 -anchor w
    entry .att.name.field -textvar name($the_node_id) -width 10 \
	-relief sunken
    frame .att.actions
    button .att.actions.ok -text "Set" \
	-command "destroy .att"
    button .att.actions.cancel -text "Cancel" \
	-command "destroy .att\n set name($the_node_id) {$name($the_node_id)}"

    pack .att.name.label .att.name.field -side left -fill x
    pack .att.actions.ok .att.actions.cancel -side left -fill x
    pack .att.name .att.actions -side top

}

######################################################################
# Edit_Arc_Attributes -
# Open the window for editing the attributes of an arc.
# Coordinates (Xpos,Ypos) are passed to the procedure. They are NOT 
# used to find the object, but to position the window close to the object.
#
# Parameters:
# - win		the window where this is happening
# - Xpos	a x coordinate
# - Ypos	a y coordinate
# - node_id	id of the link to edit
#
proc Edit_Arc_Attributes { win Xpos Ypos link_id } {
    global arc_proba arc_reward GeomParams
    global REWARDS

    if [catch {toplevel .att}] { return }

    wm title .att "Attributes of Link $link_id"

    wm geometry .att "+[expr $Xpos+$GeomParams(object_distance)/2]+$Ypos"

    frame .att.proba
    label .att.proba.label -text "Proba" -width 15 -anchor w
    entry .att.proba.field -textvar arc_proba($link_id) -width 10 \
	-relief sunken

    if $REWARDS {
	frame .att.reward
	label .att.reward.label -text "Reward" -width 15 -anchor w
	entry .att.reward.field -textvar arc_reward($link_id) -width 10 \
	    -relief sunken
	pack .att.reward.label .att.reward.field -side left
    }
	
    frame .att.actions
    button .att.actions.ok -text "Set" \
	-underline 0 \
	-command "Display_Proba $win $link_id\ndestroy .att"
    button .att.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy .att\n set arc_proba($link_id) $arc_proba($link_id)\n set arc_reward($link_id) $arc_reward($link_id)\n Display_Proba $win $link_id"
    
    pack .att.proba.label .att.proba.field -side left
    pack .att.actions.ok .att.actions.cancel -side left
    pack .att.proba -side top -anchor w -fill x
    if $REWARDS {
	pack .att.reward -side top -anchor w -fill x
    }
    pack .att.actions -side top -anchor center

    bind .att <Return> {.att.actions.ok invoke}
    bind .att <Escape> {.att.actions.cancel invoke}

#    grab .att
}

proc Start_Arc { win x y } {
    global curlinex curliney cur_arc
    global node_radius trans_width trans_height
    global orig_obj
    global DEBUG

    set curobj [$win find withtag current]
    if $DEBUG {
	puts "pointing over object $curobj"
    }

    if { $curobj == "" } { return }

    set from [$win coords current]
    set objtags  [$win gettags $curobj]

    if {  -1 != [lsearch -exact $objtags NODE] } {
	set curlinex [expr [lindex $from 0]+$node_radius]
	set curliney [expr [lindex $from 1]+$node_radius]

	$win configure -cursor cross
	if [info exists cur_arc] { unset cur_arc }
	set orig_obj $curobj
	bind $win <1> { Continue_Arc %W [%W canvasx %x] [%W canvasy %y] }
    }
}

proc Continue_Arc { win x y } {
    global curlinex curliney cur_arc 
    global node_radius arc_width create_props

    # decide if it is the end ot not.
    if [Stop_Arc $win $x $y] {
	return 
    }

    if ![info exists cur_arc] {
	# have to start the link
	set end [clip $x $y $curlinex $curliney $node_radius]
	set cur_arc [$win create line \
			 [lindex $end 0] [lindex $end 1] $x $y \
			 -tag ARC -arrow last -width $arc_width ]
    } else {
	# have to continue the existing link
	set ptlist [$win coords $cur_arc ]
	set ptlist [concat $ptlist [list $x $y]]
	eval {$win coords $cur_arc} $ptlist
    }

    set curlinex $x
    set curliney $y
}

proc Abort_Arc { win } {
    global cur_arc

    if [info exists cur_arc] {
	$win delete $cur_arc
	unset cur_arc
    }
    bind $win <1> { Start_Arc %W [%W canvasx %x] [%W canvasy %y] }
}

proc Stop_Arc { win x y } {
    global curlinex curliney cur_arc
    global node_radius arc_width arc_proba arc_reward create_props
    global orig_obj arc_list
    global DEBUG

    set cur_obj [$win find withtag current]
    if $DEBUG {
	puts "Current object = $cur_obj"
    }

    if { $cur_obj == "" } { 
	return 0
    } elseif ![Is_Type $cur_obj NODE] {
	return 0
    }

    set objtags  [$win gettags $cur_obj]
    if $DEBUG {
	puts " -> Has tags $objtags"
    }
    set goal [$win coords current]

    set end [clip $curlinex $curliney \
		 [expr [lindex $goal 0]+$node_radius] \
		 [expr [lindex $goal 1]+$node_radius] \
		 $node_radius ]
    # perhaps clipping of origin?
    if ![info exists cur_arc] {
	set start [clip $x $y $curlinex $curliney $node_radius]
	set curlinex [lindex $start 0]
	set curliney [lindex $start 1]
    }

    if ![info exists cur_arc] {
	set ptlist [list $curlinex $curliney \
			[lindex $end 0] [lindex $end 1] ]
	set cur_arc [eval {$win create line} $ptlist \
			 -tags ARC \
			 -arrow last \
			 -width $arc_width]
    } else {
	set ptlist [$win coords $cur_arc ]
	set ptlist [concat $ptlist $end]
	eval {$win coords $cur_arc} $ptlist
	eval {$win itemconfigure $cur_arc} -tags ARC
    }

    set arc_proba($cur_arc) $create_props(proba)
    set arc_reward($cur_arc) $create_props(reward)

    Display_Proba $win $cur_arc

    # update of arc lists
    lappend arc_list [list $cur_arc $orig_obj $cur_obj]
    
    $win configure -cursor ""
    bind $win <1> { Start_Arc %W [%W canvasx %x] [%W canvasy %y] }
    bind $win <2> { Abort_Arc %W }
    bind $win <3> { Exit_Mode arc }

    unset cur_arc

    return 1
}

#######################################################################
# Delete_Object -
# Removes an object from the canvas. Removes also the "attached" objects
# such as text, links.
# Handles the redirection of the attention to the "owner" of the object
# in case the object is a text.
#
# Parameters:
# - win		the canvas where this is happening
# - obj_id	the id of the object to remove with its "environment"
#
proc Delete_Object { win obj_id } {
    global node_list arc_list state
    global NUMBERS DEBUG

    if $DEBUG {
	puts -nonewline \
	    "Deleting object $obj_id, tags = [$win gettags $obj_id]"
    }

    set newnodes {}
    set newarcs {}

    if [Is_Type $obj_id NODE] {
	# find arcs ending at this node and remove them
	set newarcs {}
	foreach arc $arc_list {
	    set arc_id [lindex $arc 0]
	    set origid [lindex $arc 1]
	    set destid [lindex $arc 2]
	    if [expr { $origid == $obj_id } || { $destid == $obj_id } ] {
		Delete_Object $win $arc_id
	    } else {
		lappend newarcs $arc
	    }
	}
	# remove node itself, including from the list of nodes
	foreach node $node_list {
	    if { $obj_id != [lindex $node 0] } {
		lappend newnodes $node
	    }
	}
	set arc_list $newarcs
	set node_list $newnodes

    } elseif [Is_Type $obj_id ARC] {
	foreach arc $arc_list {
	    if { $obj_id != [lindex $arc 0] } {
		lappend newarcs $arc
	    }
	}
	set arc_list $newarcs
    } elseif [Is_Type $obj_id TEXT] {
	# might be text. Can't delete it directly. Have to find to whom it
	# belongs.
	set objtags  [$win gettags $obj_id]
	foreach tag $objtags {
	    if [expr [string match "Tx*" $tag] \
		    || [string match "Pr*" $tag] \
		    || [string match "Re*" $tag] ] {
		set new_obj_id [string range $tag 2 end]
	    }
	}
	if [info exists new_obj_id] {
	    Delete_Object $win $new_obj_id
	}
	# whatever the result, return.
	return
    }

    # remove the object from the canvas
    $win delete $obj_id
    # and its text too, if any
    set textid [$win find withtag "Tx$obj_id"]
    $win delete $textid
    set textid [$win find withtag "Pr$obj_id"]
    $win delete $textid
    set textid [$win find withtag "Re$obj_id"]
    $win delete $textid
    set selid [$win find withtag "Sel$obj_id"]
    $win delete $selid

    # renumber notes if needed
    if $NUMBERS {
	Update_Numbers $win
    }

    if $DEBUG {
	puts " ...done."
    }

    # if graph empty, restore state
    if [expr [llength $node_list] == 0 ] {
	set state($win) saved
    } else {
	set state($win) modified
    }
}

#######################################################################
# Delete_Current_Object -
# Removes the "current" object, usually the one pointed at. 
#
# Parameters:
# - win		the canvas where this is happening
#
proc Delete_Current_Object { win } {
    global DEBUG 

    set cur_obj [$win find withtag current]
    if { [llength $cur_obj] > 1 } { 
	puts "Warning! More than one current object!"
    }
    if { [llength $cur_obj] > 0 } { 
	Delete_Object $win [lindex $cur_obj 0]
    }
}
