#
# Module containing procedures related to copying objects on the canvas.
#
########################################################################
# Copy_Object -
# A procedure to copy an object with some offset (dx,dy).
# The procedure takes care of copying the attributes of the object
# according to its type.
# Associated objects are treated as follows:
# marks and texts are not moved, but the attribute is copied, and the
# Display procedure is called on the new obj.
#
# Parameters:
# - win		the canvas where the objects are
# - obj_id	the id of the object
# - dx		offset in the x direction
# - dy		offset in the y direction
#
proc Copy_Object { win obj_id dx dy } {
    global name mark PlaceTime TransTime link_weight
    global curzoomvalue place_radius trans_width trans_height
    global Copy_Table
    global DEBUG

    set coords [$win coords $obj_id]
    set new_id ""
    set scale [expr $curzoomvalue/100.0]

    if [Is_Type $obj_id PLACE] {

	set new_id [Make_Place $win \
			[list \
			     -x [expr [lindex $coords 0]+$place_radius*$scale] \
			     -y [expr [lindex $coords 1]+$place_radius*$scale] \
			     -marking $mark($obj_id) \
			     -lawparams $PlaceTime($obj_id) \
			     -name $name($obj_id) \
			     ] ]

	if { $new_id != "" } {
	    Move_Object $win $new_id $dx $dy
	}

    } elseif [Is_Type $obj_id TRANS] {

	set new_id [Make_Trans $win \
			[list \
			     -x [expr [lindex $coords 0]+$trans_width*$scale] \
			     -y [expr [lindex $coords 1]+$trans_height*$scale] \
			     -lawparams $TransTime($obj_id) \
			     -name $name($obj_id) \
			    ] ]

	if { $new_id != "" } {
	    Move_Object $win $new_id $dx $dy
	}

    } elseif [Is_Type $obj_id LINK] {
	# In order to use the Put_TraPla_Link and Put_PlaTra_Link procs
	# I need to:
	# -find out if it is trapla or platra
	# -find out the ids of the NEW extremities
	# -chop off extremities of original coordinates.
	# -shift original coordinates.
	# The resuling object has its final coords. No move is necessary
	set props [Get_Link_Props $obj_id]

	set type   [lindex $props 0]
	set fromid $Copy_Table([lindex $props 1])
	set toid   $Copy_Table([lindex $props 2])

	set coords [lrange $coords 2 end]
	set idx [expr [llength $coords]-3]
	set coords [lrange $coords 0 $idx]
	set coords [Shift_Line_Coords $coords $dx $dy]

	if { $type == "platra" } {
	    set new_id [Put_PlaTra_Link $win $fromid $toid $coords]
	} elseif { $type == "trapla" } {
	    set new_id [Put_TraPla_Link $win $fromid $toid $coords]
	} else {
	    puts "Error: unknown link type for object $obj_id"
	    return 
	}
	set link_weight($new_id) $link_weight($obj_id)

	Display_Weight $win $new_id

    } elseif [Is_Type $obj_id SELZONE] {

	set new_id [eval {$win create rectangle} $coords \
		       -outline pink -width 2 -tags SELZONE]
	$win move $new_id $dx $dy

    }

    return $new_id
}

proc Copy_Object_Or_Zone { win xto yto } {
    global cur_picked_obj
    global Copy_Table
    global DEBUG

    if ![info exists cur_picked_obj] { return }

    set bbox [$win bbox $cur_picked_obj]
    set dx [expr $xto-[lindex $bbox 0]]
    set dy [expr $yto-[lindex $bbox 1]]

    if $DEBUG {
	puts -nonewline "Copy+Shift of $dx $dy. "
    }

    if { [Is_Type $cur_picked_obj PLACE] \
	     || [Is_Type $cur_picked_obj SOURCE] \
	     || [Is_Type $cur_picked_obj TRANS] } {
	Copy_Object $win $cur_picked_obj $dx $dy 
    } elseif { [Is_Type $cur_picked_obj SELZONE] } {
	set targets [eval {$win find enclosed} $bbox]
	if $DEBUG {
	    puts "Copying objects $targets"
	}
	# must copy places and transitions BEFORE links in order to 
	# set up Copy_Table properly.
	foreach obj $targets {
	    if { [Is_Type $obj PLACE] || [Is_Type $obj SOURCE] \
		     || [Is_Type $obj TRANS] } {
		# puts "--> Copying object $obj"
		set Copy_Table($obj) [Copy_Object $win $obj $dx $dy]
	    }
	}
	# now can copy links
	foreach obj $targets {
	    if [Is_Type $obj LINK] {
		# puts "--> Copying object $obj"
		Copy_Object $win $obj $dx $dy 
	    }
	}

	# copy zone itself
	Copy_Object $win $cur_picked_obj $dx $dy

    } else {
	puts "Error: current object $cur_picked_obj should not be there."
    }
}

